package explo40th {
	import __AS3__.vec.Vector;
	
	import com.transmote.flar.FLARManager;
	import com.transmote.flar.marker.FLARMarker;
	import com.transmote.flar.marker.FLARMarkerEvent;
	import com.transmote.flar.source.FLARCameraSource;
	import com.transmote.flar.utils.FLARPVGeomUtils;
	import com.transmote.utils.time.Timeout;
	
	import flash.display.*;
	import flash.events.ActivityEvent;
	import flash.events.ErrorEvent;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Vector3D;
	import flash.media.Microphone;
	import flash.media.SoundTransform;
	import flash.text.TextField;
	import flash.text.TextFieldAutoSize;
	import flash.text.TextFormat;
	
	import org.libspark.flartoolkit.support.pv3d.FLARCamera3D;
	import org.papervision3d.core.math.Matrix3D;
	import org.papervision3d.lights.PointLight3D;
	import org.papervision3d.materials.BitmapMaterial;
	import org.papervision3d.objects.DisplayObject3D;
	import org.papervision3d.objects.parsers.DAE;
	import org.papervision3d.objects.primitives.Plane;
	import org.papervision3d.objects.primitives.Sphere;
	import org.papervision3d.render.LazyRenderEngine;
	import org.papervision3d.scenes.Scene3D;
	import org.papervision3d.view.Viewport3D;

	public class ExploFLARView extends Sprite {
		private static const SOUND_LEVEL_FLICKER:Number = 5;
		private static const SOUND_LEVEL_EXTINGUISH:Number = 25;
		private static const FLAME_FLICKER_TIME:Number = 1500;
		private static const FLAME_EXTINGUISH_TIME:Number = 7500;
		private static const FLAME_EXTINGUISH_CHANCE:Number = 0.5;
		
		private static const MODEL_SCALE:Number = 0.65;
		private static const NUM_FLAMES:int = 12;
		private static var FLAME_LOCATIONS:Vector.<Vector3D>;
		
		private var flarManager:FLARManager;
		private var microphone:Microphone;
		
		private var scene3D:Scene3D;
		private var camera3D:FLARCamera3D;
		private var viewport3D:Viewport3D;
		private var renderEngine:LazyRenderEngine;
		private var pointLight3D:PointLight3D;
		
		private var activeMarker:FLARMarker;
		private var modelContainer:DisplayObject3D;
		
		[Embed(source="../resources/assets/assets.swf", symbol="flame")]
		private var FlameTextureSymbol:Class;
		
		[Embed(source="../resources/assets/assets.swf", symbol="smoke")]
		private var SmokeTextureSymbol:Class;
		
		//private var preview:Sprite;
		
		private var flameContainer:DisplayObject3D;
		private var flameTextures:Array;
		private var flamePlanes:Array;
		private var flameGfxs:Array;
		private var flameTarget:Sphere;
		private var smokeTextures:Array;
		
		private var bFlickering:Boolean;
		private var flickerTimeout:Timeout;
		private var bExtinguishing:Boolean;
		private var bExtinguished:Boolean;
		private var extinguishStartTime:Number;
		private var extinguishingByFlame:Vector.<Boolean>;
		
		// added by David Beck to manage error reporting
		private var webcam:Boolean = true;
		private var errorText:Sprite;
		private var errorTimeout:Timeout;

		
		public function ExploFLARView () {
			this.initFLAR();
			this.initMicrophone();
			
			this.addEventListener(MouseEvent.CLICK, this.onTempClick);
		}
		
		private function onTempClick (evt:MouseEvent) :void {
			this.extinguishFlames();
		}
		
		// Creates a text box displaying the error from FLARCameraSource
		private function displayError($evt:ErrorEvent):void
		{
			webcam = false;
			errorText = new Sprite();
			var ltext:TextField = new TextField();
			var ftext:TextFormat = new TextFormat();
			ftext.font = "Arial";
			ltext.defaultTextFormat = ftext;
			ltext.textColor = 0x000000;
			ltext.autoSize = TextFieldAutoSize.LEFT;
			ltext.width = 180;
			ltext.wordWrap = true;
			ltext.text = $evt.text;
			ltext.x = 10;
			ltext.y = 10;
			errorText.graphics.beginFill(0xFFFFFF,1);
			errorText.graphics.lineStyle(2,0xfd6800,1);
			errorText.graphics.moveTo(0,0);
			errorText.graphics.lineTo(200,0);
			errorText.graphics.lineTo(200,ltext.height+20);
			errorText.graphics.lineTo(0,ltext.height+20);
			errorText.graphics.lineTo(0,0);
			errorText.addChild(ltext);
			errorText.x = 220;
			errorText.y = 159;
			this.addChild(errorText);
		}
		
		private function initFLAR () :void {
			// pass the path to the FLARManager xml config file into the FLARManager constructor.
			// FLARManager creates and uses a FLARCameraSource by default.
			// the image from the first detected camera will be used for marker detection.
			this.flarManager = new FLARManager();
			//FLARCameraSource(this.flarManager.flarSource).cameraIndex = 2;
			
			// pass the path to the FLARManager config file into FLARManager.initFromFile.
			this.flarManager.initFromFile("../resources/flar/flarConfig.xml");
			
			// add FLARManager.flarSource to the display list to display the video capture.
			this.addChild(Sprite(this.flarManager.flarSource));
			
			// listen for errors from FLARCameraSource --- added by David Beck
			FLARCameraSource(this.flarManager.flarSource).addEventListener(ErrorEvent.ERROR, this.displayError);
			
			// begin listening for FLARMarkerEvents.
			this.flarManager.addEventListener(FLARMarkerEvent.MARKER_ADDED, this.onMarkerAdded);
			this.flarManager.addEventListener(FLARMarkerEvent.MARKER_UPDATED, this.onMarkerUpdated);
			this.flarManager.addEventListener(FLARMarkerEvent.MARKER_REMOVED, this.onMarkerRemoved);
			
			// wait for FLARManager to initialize before setting up Papervision3D environment.
			this.flarManager.addEventListener(Event.INIT, this.onFlarManagerInited);
		}
		
		private function initMicrophone () :void {
			this.microphone = Microphone.getMicrophone();
			this.microphone.setLoopBack(true);
			this.microphone.setUseEchoSuppression(true);
			this.microphone.setSilenceLevel(5, 500);
			this.microphone.soundTransform = new SoundTransform(0);
			this.microphone.addEventListener(ActivityEvent.ACTIVITY, this.onMicrophoneActivity);
		}
		
		private function onFlarManagerInited (evt:Event) :void {
			if (webcam) this.initMicrophone();
			this.flarManager.removeEventListener(Event.INIT, this.onFlarManagerInited);
			
			this.scene3D = new Scene3D();
			
			// initialize FLARCamera3D with parsed camera parameters.
			this.camera3D = new FLARCamera3D(this.flarManager.cameraParams);
			
			this.viewport3D = new Viewport3D(this.stage.stageWidth, this.stage.stageHeight);
			this.addChild(this.viewport3D);
			
			// flip display to match mirrored source;
			// this is temporary until FLARManager has full support for mirroring.
			this.viewport3D.x = this.stage.stageWidth;
			this.viewport3D.scaleX = -1;
			
			this.renderEngine = new LazyRenderEngine(this.scene3D, this.camera3D, this.viewport3D);
			
			this.pointLight3D = new PointLight3D();
			this.pointLight3D.x = 1000;
			this.pointLight3D.y = 1000;
			this.pointLight3D.z = -1000;
			
			// load the model.
			// (this model has to be scaled and rotated to fit the marker; every model is different.)
			var model:DAE = new DAE(true, "model", true);
			model.load("../resources/assets/cake9/cake9.dae");
			model.rotationZ = 180;
			model.scale = MODEL_SCALE;
			
			// create a container for the model, that will accept matrix transformations.
			this.modelContainer = new DisplayObject3D();
			this.modelContainer.addChild(model);
			this.modelContainer.visible = false;
			this.scene3D.addChild(this.modelContainer);
			
			this.createCandleFlames();
			
			this.addEventListener(Event.ENTER_FRAME, this.onEnterFrame);
		}
		
		private function onMarkerAdded (evt:FLARMarkerEvent) :void {
			this.modelContainer.visible = this.flameContainer.visible = true;
			this.activeMarker = evt.marker;
			
			// reset candles if extinguished
			if (this.bExtinguished) {
				this.bExtinguishing = this.bExtinguished = false;
			}
		}
		
		private function onMarkerUpdated (evt:FLARMarkerEvent) :void {
			this.modelContainer.visible = this.flameContainer.visible = true;
			this.activeMarker = evt.marker;
		}
		
		private function onMarkerRemoved (evt:FLARMarkerEvent) :void {
			this.modelContainer.visible = this.flameContainer.visible = false;
			this.activeMarker = null;
		}
		
		private function onEnterFrame (evt:Event) :void {
			// apply the FLARToolkit transformation matrix to the model and flames.
			if (this.activeMarker) {
				var transmat:Matrix3D = FLARPVGeomUtils.convertFLARMatrixToPVMatrix(this.activeMarker.transformMatrix);
				
				// tilt cake to match table graphic
				//transmat.n31 += 0.2;
				
				this.modelContainer.transform = transmat;
				this.flameContainer.transform = transmat;
				
				this.updateFlames();
			}
			
			// update the Papervision3D view.
			this.renderEngine.render();
		}
		
		private function updateFlames () :void {
			if (!this.flamePlanes) { return; }
			
			// update flame textures randomly ('flicker'), and point flame planes at camera.
			var i:int = NUM_FLAMES;
			var flamePlane:DisplayObject3D;
			var flameGfx:Plane;
			
			var transmat:Matrix3D = FLARPVGeomUtils.convertFLARMatrixToPVMatrix(this.activeMarker.transformMatrix);
			transmat.invert();
			var flameTransmat:Matrix3D = new Matrix3D();
			
			var extinguishCounter:Number;
			if (this.bExtinguishing) {
				if (this.bExtinguished) {
					extinguishCounter = 0;
				} else {
					extinguishCounter = (FLAME_EXTINGUISH_TIME - (new Date().getTime() - this.extinguishStartTime)) / FLAME_EXTINGUISH_TIME;
				}
				
				if (extinguishCounter <= 0) {
					this.bExtinguished = true;
				} else {
					extinguishCounter = Math.pow(extinguishCounter, 2);
				}
			}
			
			while (i--) {
				flamePlane = this.flamePlanes[i];
				
				// PV3D seems to manipulate Matrix3D instances after they're stored in a DisplayObject3D.transform;
				// therefore, we have to create a new Matrix3D to use for each flame.
				flameTransmat = new Matrix3D();
				flameTransmat.copy(transmat);
				
				flameTransmat.n14 = flamePlane.x;
				flameTransmat.n24 = flamePlane.y;
				flameTransmat.n34 = flamePlane.z;
				flamePlane.transform = flameTransmat;
				
				flameGfx = this.flameGfxs[i];
				if (this.bExtinguishing) {
					// if extinguishing, randomly extinguish each candle individually
					if (!this.extinguishingByFlame[i]) {
						// if this candle is not yet extinguished, there is a chance it will become extinguished.
						// until then, it flickers.
						if (Math.random() < FLAME_EXTINGUISH_CHANCE) {
							this.extinguishingByFlame[i] = true;
						}
						this.flickerFlame(flamePlane);
					} else {
						// if this candle is extinguished, randomly select a frame from the smoke animation
						flameGfx.material = this.smokeTextures[Math.floor(Math.random()*this.smokeTextures.length)];
						flameGfx.alpha = 0.25 + 0.75 * extinguishCounter;
						flameGfx.scaleY = 0.75 + (1-extinguishCounter);
					}
				} else {
					// if not extinguishing, randomly select a frame from the flame animation
					flameGfx.material = this.flameTextures[Math.floor(Math.random()*this.flameTextures.length)];
					flameGfx.alpha = 1.0;
					flameGfx.scaleY = 1.0;
				}
				
				if (this.bFlickering) {
					this.flickerFlame(flamePlane);
				}
			}
		}
		
		private function flickerFlame (flamePlane:DisplayObject3D) :void {
			flamePlane.rotationX = Math.random() * 60;
			flamePlane.rotationZ = -90 + Math.random() * 45;
			flamePlane.scaleY = 0.6 + Math.random() * 0.8;
		}
		
		private function createCandleFlames () :void {
			this.loadFlameTextures();
			this.flamePlanes = new Array(NUM_FLAMES);
			this.flameGfxs = new Array(NUM_FLAMES);
			this.flameContainer = new DisplayObject3D();
			this.scene3D.addChild(this.flameContainer);
			
			FLAME_LOCATIONS = new Vector.<Vector3D>(NUM_FLAMES, true);
			FLAME_LOCATIONS[0] = new Vector3D(43.329, 62.338, 125);
			FLAME_LOCATIONS[1] = new Vector3D(30.65, 48.507, 125);
			FLAME_LOCATIONS[2] = new Vector3D(0.271, 77.039, 125);
			FLAME_LOCATIONS[3] = new Vector3D(-46.091, 46.713, 125);
			FLAME_LOCATIONS[4] = new Vector3D(-61.88, 28.355, 125);
			FLAME_LOCATIONS[5] = new Vector3D(-77.551, 5.592, 125);
			FLAME_LOCATIONS[6] = new Vector3D(-45.821, -19.29, 125);
			FLAME_LOCATIONS[7] = new Vector3D(-46.603, -60.42, 125);
			FLAME_LOCATIONS[8] = new Vector3D(-26.193, -27.186, 125);
			FLAME_LOCATIONS[9] = new Vector3D(42.545, -54.109, 125);
			FLAME_LOCATIONS[10] = new Vector3D(42.268, -18.869, 125);
			FLAME_LOCATIONS[11] = new Vector3D(70.177, -22.518, 125);
			
			const FLAME_GFX_OFFSET_Y:Number = 10;
			
			var i:int = NUM_FLAMES;
			var flamePlane:DisplayObject3D;
			var flameGfx:Plane;
			while (i--) {
				flameGfx = new Plane(null, 10, 20);
				flameGfx.material = this.flameTextures[Math.floor(Math.random()*this.flameTextures.length)];
				flameGfx.y = FLAME_GFX_OFFSET_Y;
				flameGfx.useOwnContainer = true;
				this.flameGfxs[i] = flameGfx;
				
				flamePlane = new DisplayObject3D();
				flamePlane.addChild(flameGfx);
				
				flamePlane.x = -FLAME_LOCATIONS[i].x * MODEL_SCALE;
				flamePlane.y = FLAME_LOCATIONS[i].y * MODEL_SCALE;
				flamePlane.z = FLAME_LOCATIONS[i].z * MODEL_SCALE;
				
				this.flameContainer.addChild(flamePlane);
				
				this.flamePlanes[i] = flamePlane;
				
				/*
				// TEMP FOR TESTING
				var originMarker:Sphere = new Sphere(new FlatShadeMaterial(this.pointLight3D, 0xFF1919, 0x730000), 10, 4, 4);
				originMarker.scale = 0.1;
				flamePlane.addChild(originMarker);
				*/
			}
			this.flameContainer.visible = false;
		}
		
		private function loadFlameTextures () :void {
			var flameTextureMC:MovieClip = new FlameTextureSymbol();
			var flameSnapshot:BitmapData;
			var flameTexture:BitmapMaterial;
			this.flameTextures = new Array(flameTextureMC.totalFrames);
			for (var i:int=1; i<=flameTextureMC.totalFrames; i++) {
				flameTextureMC.gotoAndStop(i);
				flameSnapshot = new BitmapData(flameTextureMC.width, flameTextureMC.height, true, 0);
				flameSnapshot.draw(flameTextureMC);
				flameTexture = new BitmapMaterial(flameSnapshot);
				flameTexture.doubleSided = true;
				this.flameTextures[i-1] = flameTexture;
			}
			
			var smokeTextureMC:MovieClip = new SmokeTextureSymbol();
			var smokeSnapshot:BitmapData;
			var smokeTexture:BitmapMaterial;
			this.smokeTextures = new Array(smokeTextureMC.totalFrames);
			for (i=1; i<=smokeTextureMC.totalFrames; i++) {
				smokeTextureMC.gotoAndStop(i);
				smokeSnapshot = new BitmapData(smokeTextureMC.width, smokeTextureMC.height, true, 0);
				smokeSnapshot.draw(smokeTextureMC);
				smokeTexture = new BitmapMaterial(smokeSnapshot);
				smokeTexture.doubleSided = true;
				this.smokeTextures[i-1] = smokeTexture;
			}
		}
		
		private function onMicrophoneActivity (evt:ActivityEvent) :void {
			//trace("mic activity:"+this.microphone.activityLevel);
			
			if (this.microphone.activityLevel > SOUND_LEVEL_EXTINGUISH) {
				this.extinguishFlames();
			} else if (this.microphone.activityLevel > SOUND_LEVEL_FLICKER) {
				this.flickerFlames();
			}
		}
		
		private function flickerFlames () :void {
			if (this.bFlickering || this.bExtinguishing || this.bExtinguished) { return; }
			
			this.bFlickering = true;
			this.flickerTimeout = new Timeout(this.onFlickerComplete, FLAME_FLICKER_TIME);
		}
		
		private function onFlickerComplete () :void {
			this.bFlickering = false;
		}
		
		private function extinguishFlames () :void {
			this.bFlickering = false;
			this.bExtinguishing = true;
			this.extinguishStartTime = new Date().getTime();
			this.extinguishingByFlame = new Vector.<Boolean>(NUM_FLAMES, true);
		}
	}
}